


rawset(_G, "DITTO_CONFIG", {
	KEEP_COLOR= 1,
	DONT_START_RANDOM= 2,
	TRANFORM_ONLY_ONCE= 4,
    BATTLE_COPY_DEADPLAYERS= 8,
	KEEP_HM_RESTAT= 16,
	FRIENDMOD_IGNORE_TEAM_COLOR= 32,


    INITIAL_CHECK= 2048
})

-- this particular config should only be handled by the server, I decided.
local always_server_locked= DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS

-- used for debugging scenarios, did its job, leaving it here, free to chill
local function DITTO_log(msg)
	print("[DITTO] "..msg)
end

-- print the current state of a setting
-- (only single config state, no combination)
-- pov depends on serv boolean arg (if true server config pov,
-- else player's)
local function _playerDittoConfigStateStatusString(p, state, serv)
    if serv==nil then serv= false end
    local conf= p.ditto_config
	if serv then conf= server.ditto_server_config.base end
	
    -- if state is lock by server, then it the state is how the serv say so
    -- either, it's the one chosen by player
    local b= ((~server.ditto_server_config.lockMask) & conf) 
                | (server.ditto_server_config.lockMask & server.ditto_server_config.base)

    -- is this the state always controlled by server?
    local spe= (state & always_server_locked)
    return 
        ((b & state) and "\131On \128" or "\130Off\128")
        ..((server.ditto_server_config.lockMask & state) and " \133[ "..(spe and "SERVER" or "LOCKED").." ]\128" or "\128")
end

-- print a config state,
-- either the server's (it 'serv==true'), or player's ('serv==false')
local function printPlayerDittoConfig(p, serv, current)
    if serv==nil then serv= false end
    current= ($==nil or $) and 'current' or 'new'

    if (not p) or (p.ditto_config==nil) then
        print("\133Ditto - Error\128: player ditto data missing or not initialized yet?!")

        return
    end


    CONS_Printf(p,"----------------------------")
    CONS_Printf(p,(serv and "\130SERVER's" or "\129Your").."\128 "..current.." 'Ditto' config:")
    CONS_Printf(p, "    Keep base color [ \140keep_color\128 ]:                    "
        .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.KEEP_COLOR, serv))
    CONS_Printf(p, "    Don't start at random [ \140no_random\128 ]:               "
        .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.DONT_START_RANDOM, serv))
    CONS_Printf(p, "    Transform only once [ \140copy_once\128 ]:                 "
        .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.TRANFORM_ONLY_ONCE, serv))
    -- this one is server specific
    if serv then
        CONS_Printf(p, "    Battle copy ghost players [ \140battle_dead_copy\128 ]:    "
            .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS, serv))
    end
    -- these following ones are targeting external mods
	local cv_restat= CV_FindVar("hm_restat")
    if cv_restat and cv_restat.value then
        CONS_Printf(p, "    Keep own's restat change [ \140keep_restat\128 ]:          "
            .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.KEEP_HM_RESTAT, serv))
    end
    if FRIENDMOD_CheckTeams then
        CONS_Printf(p, "    Ignore friendmod's team color [ \140ignore_team\128 ]:     "
            .._playerDittoConfigStateStatusString(p, DITTO_CONFIG.FRIENDMOD_IGNORE_TEAM_COLOR, serv))
    end
    CONS_Printf(p,"----------------------------")

    return
end



-- this function changes either the server's config (it 'serv==true'),
-- either the player's 'serv==false', according the strings arguments
-- passed. Ex:
--    'keep_color' toggles the 'keep_color' setting
--    'no_random off' turns off the 'no_random' setting
--    'copy_once on' turns on the 'copy_once' setting
-- If no setting string is given, only show current states
-- On server POV:
--    'keep_color lock' locks the current server states for all players
--    'keep_color unlock' unlocks it
local function _dittoConfigManage(p, serv, ...)
    local args = {...}

	if not ... then 
        printPlayerDittoConfig(p, serv)

        return
	end

	local config_change= 0
    local config_switch= -1
	for i=1,#args do
		local arg= args[i]:gsub("'", '"'):lower()

        local num= tonumber(arg)
        -- if we encounter "on", "off", or "toggle" strings, we respectively
        -- turn on, turn off or toggle the settings corresponding to the settings strings
        -- given before hand
        if (num~=nil or arg=="on" or arg=="off" or arg=="toggle") then
            config_switch= ( ((num~=nil and num>=1) or arg=="on")
                                and 1
                                or ( ((num~=nil and num<0) or arg=="toggle")
                                    and (-1)
                                    or 0
                                )
                            )
            if config_change~=0 then
                local r= p.ditto_config
				if serv then r= server.ditto_server_config.base end
				
				if (config_switch<0) then r= ($ | config_change) & (~($ & config_change))
				elseif (config_switch>0) then r= ( $ | config_change )
				else r= ( $ & (~config_change)) end

                if serv then
                    server.ditto_server_config.base= r
                else
                    p.ditto_config= ((~server.ditto_server_config.lockMask) & r) 
                                | (server.ditto_server_config.lockMask & server.ditto_server_config.base)
                end
                -- previous settings have been delt with, flush them out
                config_change= 0
            end
        -- lock or unlock previously given settings (server pov only)
        elseif serv and arg=="lock" then
            -- sets the 'lock mask' that says which states are locked on the player pov
            server.ditto_server_config.lockMask= ($ | config_change) | always_server_locked
            config_change= 0
        elseif serv and arg=="unlock" then
            server.ditto_server_config.lockMask= ($ & (~config_change)) | always_server_locked
            config_change= 0
		elseif arg=="color" or arg=="keep_color" then
			config_change= $ | DITTO_CONFIG.KEEP_COLOR
		elseif arg=="no_random" or arg=="no_random_start" then
			config_change= $ | DITTO_CONFIG.DONT_START_RANDOM
		elseif arg=="transform_once" or arg=="only_once" or arg=="once" or arg=="copy_once" then
			config_change= $ | DITTO_CONFIG.TRANFORM_ONLY_ONCE
        elseif arg=="battle_dead_copy" then
            if serv then
                config_change= $ | DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS
            else
                CONS_Printf(p,"\133Ditto - Error\128: Only server can change '\140battle_dead_copy\128'...")

                return
            end
		elseif arg=="keep_restat" or arg=="restat" then
			config_change= $ | DITTO_CONFIG.KEEP_HM_RESTAT
		elseif arg=="no_friend" or arg=="ignore_team" or
			arg=="ignore_team_color" or arg=="no_team_color"
		then
			config_change= $ | DITTO_CONFIG.FRIENDMOD_IGNORE_TEAM_COLOR
		else
			CONS_Printf(p,"\133Ditto - Error\128: unknown ditto setting '\140"..arg.."\128'...")

            return
		end
	end

    -- if the string args didn't end with "on", "off", "toggle", "unlock", or "lock", means that
    -- we are toggling all the settings designated beforehand
    if config_change~=0 then
        local r= p.ditto_config
        if serv then r= server.ditto_server_config.base end
        r= ($ | config_change) & (~($ & config_change))

		if serv then
            server.ditto_server_config.base= r
        else
            -- on player's pov
            -- we take into accound the 'lock mask' provided by the server that determines
            -- which state are lock (i.e. untounchable atm)
            p.ditto_config= ((~server.ditto_server_config.lockMask) & r) 
                        | (server.ditto_server_config.lockMask & server.ditto_server_config.base)
        end
    end

    -- finaly print current state
    printPlayerDittoConfig(p, serv, false)
end

COM_AddCommand("ditto_config", function(p, ...)
    if (not server) or (not p) then
        CONS_Printf(p,"\133Ditto - Error\128: server or player nil value - not initialized yet?")
        return
    end 
    if server.ditto_server_config==nil then server.ditto_server_config= {base= 0, lockMask= DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS} end
    if p.ditto_config==nil then p.ditto_config= 0 end

    _dittoConfigManage(p, false,...)
end)

COM_AddCommand("ditto_server_config", function(p, ...)
    if (not server) or (not p) then
        CONS_Printf(p,"\133Ditto - Error\128: server or player nil value - not initialized yet?")
        return
    end 
    if server.ditto_server_config==nil then server.ditto_server_config= {base= 0, lockMask= DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS} end
    if p.ditto_config==nil then p.ditto_config= 0 end

    local old_serv_config= {server.ditto_server_config.base, server.ditto_server_config.lockMask}

    _dittoConfigManage(p, true, ...)
    
    -- if there's been change on server pov, we make sure it affects the player's pov as well
    -- since servers says which state are locked and whatnot
    if (server.ditto_server_config.lockMask~=old_serv_config.lockMask) or
		(server.ditto_server_config.base~=old_serv_config.base)
	then
        for _p in players.iterate do
            _p.ditto_config= ($~=nil) and $ or 0
            
            _p.ditto_config= ((~server.ditto_server_config.lockMask) & $) 
                               | (server.ditto_server_config.lockMask & server.ditto_server_config.base)
        end
    end
end, 1)

local function _init()
    if server and not server.ditto_server_config then
        server.ditto_server_config= {base= 0, lockMask= DITTO_CONFIG.BATTLE_COPY_DEADPLAYERS}
    end

    for player in players.iterate do
        if (not player.ditto_config) or not (player.ditto_config & DITTO_CONFIG.INITIAL_CHECK) then
            player.ditto_config= server.ditto_server_config.base | DITTO_CONFIG.INITIAL_CHECK
        end
    end
end

-- ye, stuff needs to initialized asap, y'know?
addHook("IntermissionThinker", _init)
addHook("VoteThinker", _init)
addHook("MapLoad", _init)
addHook("ThinkFrame", _init)
